/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "TimeTrialTrack.h"
#include "Settings.h"
#include "Surface.h"
#include "str.h"
#include "BinaryFile.h"
#include "StateManager.h"
#include <sstream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <algorithm>
#include <ctime>
#include <cstdio>
#include <climits>

const int TIME_BITS  = 22;
const int GHOST_BITS = 17;
const std::size_t MAX_TIMES = 8;

static std::string id2filename(Uint32 id)
{
	std::string filename;
	while (id > 0)
	{
		int i = id % 36;

		if (i < 10)
		{
			filename += '0' + i;
		}
		else
		{
			filename += 'a' + (i - 10);
		}
		id /= 36;
	}
	
	// Makes file order consistent with id. 
	std::reverse(filename.begin(), filename.end());
	
	return filename;
}

static Uint32 filename2id(std::string filename)
{
	Uint32 id = 0;;
	for (std::string::const_iterator it = filename.begin(); it != filename.end(); ++it)
	{
		id *= 36;
		char c = std::tolower(*it);
		if (c >= '0' && c <= '9')
		{
			id += c - '0';
		}
		else if (c >= 'a' && c <= 'z')
		{
			id += c - 'a' + 10;
		}
	}
	
	return id;
}

static std::string name2filename(const std::string& name)
{
	std::ostringstream filename;
	for (std::string::const_iterator it = name.begin(); it != name.end(); ++it)
	{
		char c = std::tolower(*it);
		if ((c >= '0' && c <= '9') || 
		    (c >= 'a' && c <= 'z') ||
		     c == '-' || c == '_')
		{
			filename << c;
		}
		else
		{
			
			filename << '%' << std::setw(2) << std::hex << std::setfill('0') << int(c);
		}
	}
	return filename.str();
}

static int angle2int(double angle, int limit)
{
	int value = static_cast<int>((limit * angle) / PI_TWO) % limit;
	if (value < 0)
	{
		value += limit;
	}
	return value;
}

static double int2angle(int value, int limit)
{
	double angle = (value * PI_TWO) / limit;
	if (angle > PI)
	{
		// This shouldn't really be needed but if we don't do it there seems
		// to be a small rounding error that is noticeable at the start line.
		angle -= PI_TWO;
	}
	return angle;
}

TimeTrialTrack::TimeTrialTrack()
:	id(0),
	laps(-1),
	personalBest(0),
	newTimeIndex(-1),
	newTime(0),
	driverGhost(1),
	bestGhost(0)
	
{
}

TimeTrialTrack::TimeTrialTrack(const Map& map, int laps)
:	id(0),
	map(map),
	laps(laps),
	personalBest(0),
	newTimeIndex(-1),
	newTime(0),
	driverGhost(1),
	bestGhost(0)
{
}

TimeTrialTrack::TimeTrialTrack(Uint32 id)
:	id(id),
	newTimeIndex(-1),
	newTime(0),
	driverGhost(1),
	bestGhost(0)
{
	std::string trackFilename = id2filename(id);
	std::string driverFilename = name2filename(Settings::timetrial.name);
	
	Path trackPath(configPath() + ("timetrial/" + trackFilename));
	InputBinaryFile trackFile(trackPath);
	map.read(trackFile);
	laps = trackFile.read(4);
	
	int topTimesCount = trackFile.read(4);
	
	for (int i = 0; i < topTimesCount; ++i)
	{
		std::string name = trackFile.readS();
		int time = trackFile.read(TIME_BITS);
		topTimes.push_back(std::make_pair(name, time));
	}

	if (topTimes.size() > MAX_TIMES)
	{
		topTimes.resize(MAX_TIMES);
	}
	
	Path driverPath(configPath() + ("timetrial/" + driverFilename + "/" + trackFilename));
	InputBinaryFile driverFile(driverPath);
	
	oldPersonalBest = personalBest = driverFile.read(TIME_BITS);
}

void TimeTrialTrack::drawTopTimes(Surface& surface, bool compact) const
{	
	int y = 60;
	int ySpacing;
	switch (topTimes.size())
	{
	case 8:
		ySpacing = compact ? 13 : 15;
		break;
	case 7:
		ySpacing = compact ? 14 : 17;
		break;
	case 6:
		ySpacing = compact ? 17 : 20;
		break;
	default:
		ySpacing = 20;
	}
	for (std::size_t i = 0; i < topTimes.size(); ++i)
	{
		Color color = TEXT_COLOR;
		if (newTimeIndex == i)
		{
			color = TIME_STRONG_HIGHLIGHT_COLOR;
		}
		else if (topTimes[i].first == Settings::timetrial.name)
		{
			color = TIME_WEAK_HIGHLIGHT_COLOR;
		}
		Surface(topTimes[i].first, BIG_FONT, color, TEXT_BORDER_COLOR).draw(surface, 96, y);
		Surface valueSurface(time2string(topTimes[i].second, true), BIG_FONT, color, TEXT_BORDER_COLOR);
		valueSurface.draw(surface, 201 - valueSurface.width() / 2, y);
		y += ySpacing;
	}
	
	if (personalBest > 0)
	{
		std::string bestTimeText;
		Color color = TIME_STRONG_HIGHLIGHT_COLOR;
		if (newTimeIndex == 0)
		{
			bestTimeText = "your time " + time2string(newTime, true) + " is a new record!";
		}
		else if (newTime != 0 && (oldPersonalBest == 0 || newTime < oldPersonalBest))
		{
			bestTimeText = "your time " + time2string(newTime, true) + " is a new personal best!";
			if (newTimeIndex >= topTimes.size())
			{
				color = TIME_WEAK_HIGHLIGHT_COLOR;
			}
		}
		else if (newTime > 0)
		{
			bestTimeText = "your time is " + time2string(newTime, true);
			if (newTimeIndex >= topTimes.size())
			{
				color = TIME_WEAK_HIGHLIGHT_COLOR;
			}
		}
		else
		{
			bestTimeText = "your best time is " + time2string(personalBest, true);
			color = TIME_WEAK_HIGHLIGHT_COLOR;
		}
		Surface bestTimeSurface(bestTimeText, BIG_FONT, color, TEXT_BORDER_COLOR);
		bestTimeSurface.draw(surface, (GAME_WIDTH - bestTimeSurface.width()) / 2, compact ? 170 : 190);
	}
}

void TimeTrialTrack::addTime(int time)
{
	if (time == INT_MAX)
	{
		newTimeIndex = -1;
		newTime = 0;
		return;
	}
	
	std::string trackFilename(id2filename(id));
	Path trackPath(configPath() + ("timetrial/" + trackFilename));
	if (id == 0)
	{
		id = std::time(0);
		do
		{
			++id;
			trackFilename = id2filename(id);
			trackPath = configPath() + ("timetrial/" + trackFilename);
		} while (trackPath.exist());
	}
	
	if (personalBest == 0 || time < personalBest)
	{
		oldPersonalBest = personalBest;
		personalBest = time;
		
		std::string driverFilename = name2filename(Settings::timetrial.name);
		Path driverPath(configPath() + ("timetrial/" + driverFilename + "/" + trackFilename));
		driverPath.parent().mkdir();
		OutputBinaryFile driverFile(driverPath);
		driverFile.write(TIME_BITS, personalBest);
		
		driverFile.write(GHOST_BITS, ghostData.size());
		
		if (!ghostData.empty())
		{
			GhostState& gs = ghostData[0];
			
			driverFile.write(9, gs.x);
			driverFile.write(8, gs.y);
			driverFile.write(8, angle2int(gs.rotation, 256));
		}
		
		int dx1 = 0;
		int dy1 = 0;
		int dv1 = 0;
		for (std::size_t i = 1; i < ghostData.size(); ++i)
		{
			GhostState& gs1 = ghostData[i-1];
			GhostState& gs2 = ghostData[i];
			
			int dx2 = gs2.x - gs1.x;
			int dy2 = gs2.y - gs1.y;
			int dv2 = angle2int(gs2.rotation, 256) - angle2int(gs1.rotation, 256);
			if (dv2 > 127)
			{
				dv2 = dv2 - 256;
			}
			else if (dv2 < -127)
			{
				dv2 = dv2 + 256;
			}
			
			int ddx = dx2 - dx1;
			int ddy = dy2 - dy1;
			int ddv = dv2 - dv1;
			
			while (ddx != 0)
			{
				ddx = (ddx > 0 ? -ddx : -(ddx + 1));
				driverFile.write(1, 1);
			}
			driverFile.write(1, 0);
			
			while (ddy != 0)
			{
				ddy = (ddy > 0 ? -ddy : -(ddy + 1));
				driverFile.write(1, 1);
			}
			driverFile.write(1, 0);
			
			while (ddv != 0)
			{
				ddv = (ddv > 0 ? -ddv : -(ddv + 1));
				driverFile.write(1, 1);
			}
			driverFile.write(1, 0);
			
			dx1 = dx2;
			dy1 = dy2;
			dv1 = dv2;
		}
	}
	
	std::vector<std::pair<std::string, int> >::iterator it = topTimes.begin();
	while (it != topTimes.end() && it->second <= time)
	{
		++it;
	}
	newTimeIndex = it - topTimes.begin();
	newTime = time;
	topTimes.insert(it, std::make_pair(Settings::timetrial.name, time));
	
	if (topTimes.size() > MAX_TIMES)
	{
		topTimes.resize(MAX_TIMES);
	}
	
	OutputBinaryFile trackFile(trackPath);
	map.write(trackFile);
	trackFile.write(4, laps);
	trackFile.write(4, topTimes.size());
	for (it = topTimes.begin(); it != topTimes.end(); ++it)
	{
		trackFile.writeS(it->first);
		trackFile.write(TIME_BITS, it->second);
	}
}

const Map& TimeTrialTrack::getMap() const
{
	return map;
}

int TimeTrialTrack::getLaps() const
{
	return laps;
}

std::vector<TimeTrialTrack> TimeTrialTrack::loadFromFile()
{
	std::vector<TimeTrialTrack> tracks;
	Path trackDir(configPath() + "timetrial");
	
	std::vector<std::string> files = trackDir.files();
	std::sort(files.rbegin(), files.rend());
	
	for (std::size_t i = 0; i < files.size(); ++i)
	{
		if ((trackDir + files[i]).isFile())
		{
			Uint32 id = filename2id(files[i]);
			TimeTrialTrack ttt(id);
			if (ttt)
			{
				tracks.push_back(ttt);
			}
		}
	}
	
	return tracks;
}

TimeTrialTrack::operator bool() const
{
	return map.isValid();
}

void TimeTrialTrack::removeFiles()
{
	Path timeTrialPath(configPath() + "timetrial");
	std::string trackFilename(id2filename(id));
	
	Path trackPath(timeTrialPath + trackFilename);
	std::remove(trackPath.c_str());
	
	const std::vector<std::string>& dirs = timeTrialPath.files();
	for (std::size_t i = 0; i < dirs.size(); ++i)
	{
		Path driverPath(timeTrialPath + (dirs[i] + "/" + trackFilename));
		if (driverPath.exist())
		{
			std::remove(driverPath.c_str());
		}
	}	
}

void TimeTrialTrack::prepareGhosts(GameState& gameState)
{
	if (!topTimes.empty())
	{
		std::string trackFilename(id2filename(id));
		if (Settings::timetrial.showDriverGhost || (Settings::timetrial.showBestGhost && topTimes[0].first == Settings::timetrial.name))
		{
			loadGhost(driverGhost, Settings::timetrial.name, gameState);
		}
		else
		{
			driverGhost.disable();
		}
		if (Settings::timetrial.showBestGhost && !topTimes.empty() && topTimes[0].first != Settings::timetrial.name)
		{
			loadGhost(bestGhost, topTimes[0].first, gameState);
		}
		else
		{
			bestGhost.disable();
		}
	}
	ghostData.clear();
}

void TimeTrialTrack::ghostUpdate(int x, int y, double rotation)
{
	driverGhost.update();
	bestGhost.update();
	
	GhostState state;
	state.x = x;
	state.y = y;
	state.rotation = rotation;
	ghostData.push_back(state);
}

void TimeTrialTrack::loadGhost(Ghost& ghost, const std::string& name, GameState& gameState)
{
	InputBinaryFile file(configPath() + ("timetrial/" + name2filename(name) + "/" + id2filename(id)));
	
	// Ignore time.
	file.read(TIME_BITS);
	
	ghostData.clear();
	
	GhostState state;
	std::size_t count = file.read(GHOST_BITS);
	if (count > 0)
	{
		int x = file.read(9);
		int y = file.read(8);
		int v = file.read(8);
		
		state.x = x;
		state.y = y;
		state.rotation = int2angle(v, 256);
		ghostData.push_back(state);
		
		int dx = 0;
		int dy = 0;
		int dv = 0;
		for (std::size_t i = 1; i < count; ++i)
		{
			int ddx = 0;
			while (file.read(1))
			{
				ddx = (ddx >= 0 ? -(ddx + 1) : -ddx);
			}
			
			int ddy = 0;
			while (file.read(1))
			{
				ddy = (ddy >= 0 ? -(ddy + 1) : -ddy);
			}
			
			int ddv = 0;
			while (file.read(1))
			{
				ddv = (ddv >= 0 ? -(ddv + 1) : -ddv);
			}
			
			dx += ddx;
			dy += ddy;
			dv += ddv;
			
			x += dx;
			y += dy;
			v += dv;
			
			state.x = x;
			state.y = y;
			state.rotation = int2angle(v, 256);
			ghostData.push_back(state);
		}
	}
	
	ghost.reset(ghostData, gameState);
}
